from urllib import response
from pydantic import BaseModel
import pickle
import requests
from requests.models import Response


class Solver:
    """
    Solver class connects to QpiAI-Opt's cloud server, and run the solver on the passed problem.

    Initialize Solver class with problem and access token

    :type problem: BaseModel
    :param problem: Problem object to be passed from one of the classes in problem directory

    :type url: str
    :param url: Url to access api

    :type access_token: str
    :param access_token: access token to authenticate the solver
    """
    response: Response

    def __init__(self, problem: BaseModel, url: str, access_token: str):
        self.Problem = problem
        self.data = pickle.dumps(problem)
        self.url = url
        self.access_token = access_token
        self.response = None
        self.graph = None
    def run(self):
        """
        Runs the problem on the QpiAI-Opt Solver on cloud and receives the response
        """
        self.response = requests.post(url=f"{self.url}/{self.Problem.problem_type}",
                                      headers={"access_token": self.access_token}, data=self.data)

    def get_result(self):
        """
        to fetch the result after the solver has returned the result of the submitted problem

        :return Response: json
        """
        result = {'num_nodes': self.response.json()['num_nodes'],
                  'objective': self.response.json()['objective'],
                  'time': self.response.json()['time']}
        return result

    def get_solution(self):
        return self.response.json()['solution']

    def get_qubo_graph(self):
        if self.response:
            try:
                self.graph = self.response.json()['graph']
                return self.response.json()['graph']
            except:
                return self.response
        else:
            print("No response received! Make sure to run the Solver.run() method before accessing response metadata!")
    
    def get_qubo_dict(self):
        self.qubo_dict = dict()
        if self.response:
            try:
                self.graph = self.response.json()['graph']
                if self.graph:
                    l = len(self.graph["weights"])
                    for i in range(l):
                        self.qubo_dict[(self.graph["edges"][0][i], self.graph["edges"][1][i])] = self.graph["weights"][i]
                    return self.qubo_dict
            except:
                # print(e)
                return self.response
        else:
            print("No response received! Make sure to run the Solver.run() method before accessing response metadata!")




from pydantic import BaseModel
from typing import Optional

class NumberPartitionProblem(BaseModel):
    '''class representing a Number Partition problem to feed the inputs to the solver
    :type length: int
    :param length: Length of the list of the numbers to be partitioned equally

    :type numbers: Optional[list]
    :param numbers: List of numbers to be partitioned. If not passed as an input, the problem class generates a random list of size length

    :type penalty: int
    :param penalty: penalty term used in the QUBO formulation.

    :type max_evaluations: int
    :param max_evaluations: Number of maximum evaluations to be run to search the solution.
        '''
    class Config:
        arbitrary_types_allowed = True

    length: int = 100
    numbers: Optional[list] = None
    penalty: int = 2
    max_evaluations: int = 1000
    problem_type: str = "numberPartition"



import dimod
import neal
class SolveDwave:
    """
    This class is a wrapper around dwave's solvers to solve QUBO problems. 
    This class provides four methods to solve the problem, they are
    * exact_solver
    * simulated_annealing_sampler
    * neal_simulated_annealer
    * random_sampler
    
    :type qubo: dict
    :param qubo: qubo represented in a dictionary, for example ``qubo = {(0,0):1, (0,1):-1, (1,0):1, (1,1):-1, (2,0):-3}``
    """
    def __init__(self, qubo: dict):
        self.qubo = qubo
    
    def exact_solver(self):
        """
        Returns the exact solution for the qubo. Not suitable for bigger problems as it does an exhaustive search ibn the solution space.
        :return sampled set of solutions
        """
        return dimod.ExactSolver().sample_qubo(self.qubo)
    
    def simulated_annealing_sampler(self):
        """
        This simulated annealing solver implements a straightforward approach of the simulated annealer(Only for testing and debugging locally). Not suitable for bigger problems as it is not the best implementation, instead use ``neal_simulated_annealer`` for larger problems.
        :return sampled set of solutions
        """
        return dimod.SimulatedAnnealingSampler().sample_qubo(self.qubo)
    
    def neal_simulated_annealer(self):
        """
        This is a high performant implementation of simulated annealer given by Dwave-neal.
        :return sampled set of solutions
        """
        return neal.SimulatedAnnealingSampler().sample_qubo(self.qubo)
    
    def random_sampler(self):
        """
        This methhod performs a random sampling of the solutions and returns the sample set.
        :return sampled set of solutions
        """
        return dimod.RandomSampler().sample_qubo(self.qubo)



from typing import Union, Optional
import networkx as nx
from pydantic import BaseModel


class MaximumIndependentSetProblem(BaseModel):
    '''class representing a Maximum Independent Set problem to feed the inputs to the solver
        :type num_nodes: int
        :param num_nodes: Number of nodes in the problem graph

        :type graph: Optional[Union[dict, networkx.Graph]]
        :param graph: Graph to be solved for maximum cut. If the graph is not passed the problem method generates a random graph as per given num_nodes.

        :type penalty: int
        :param penalty: penalty term used in the QUBO formulation.

        :type max_evaluations: int
        :param max_evaluations: Number of maximum evaluations to be run to search the solution.
        '''
    class Config:
        arbitrary_types_allowed = True
    num_nodes: int = 100
    graph: Optional[Union[dict, nx.Graph]] = None
    penalty: int = 2
    max_evaluations: int = 1000
    problem_type: str = "maximumIndependentSet"


from qpiai_opt.problem.max_cut import MaximumCutProblem
from qpiai_opt.problem.vertex_cover import MinimumVertexCoverProblem
from qpiai_opt.problem.max_independet_set import MaximumIndependentSetProblem
from qpiai_opt.problem.qubo import QUBOProblem
from qpiai_opt.problem.number_partition import NumberPartitionProblem


from qiskit.utils import algorithm_globals, QuantumInstance
from qiskit.algorithms import QAOA
from qiskit import BasicAer
from qiskit_optimization import QuadraticProgram
from qiskit_optimization.algorithms import MinimumEigenOptimizer, CplexOptimizer

class SolveQiskit:
    """
    This class is a wrapper around qiskit's solvers to solve QUBO problems. 
    This class provides two methods to solve the problem, they are
    * qaoa
    * cplex
    
    :type qubo: dict
    :param qubo: qubo represented in a dictionary, for example ``qubo = {(0,0):1, (0,1):-1, (1,0):1, (1,1):-1, (2,0):-3}``.

    :type num_var: int
    :param num_var: number of variables in the qubo.
    """
    def __init__(self, qubo: dict, num_var: int):
        self.qubo = qubo
        self.qp = QuadraticProgram()
        for _ in range(num_var):
            self.qp.binary_var()
        self.qp.minimize(quadratic=self.qubo)
    
    def qaoa(self, quantum_instance: QuantumInstance = QuantumInstance(BasicAer.get_backend("qasm_simulator"))):
        """
        Runs the qiskits MinimumEigenOptmizer to solve the given qubo problem. Can use different backends supported by qiskit.
        
        :type quantum_instance: QuantumInstance
        :param quantum_instance: Can supply the quantum_instance to run on different backends supported by qiskit.

        :return : result containing solution and its objetive
        :rtype : dict
        """
        quantum_instance = quantum_instance
        qaoa_mes = QAOA(quantum_instance=quantum_instance)
        qaoa = MinimumEigenOptimizer(qaoa_mes)
        qaoa_result = qaoa.solve(self.qp)
        return {"solution":[qaoa_result.x], "objective":[qaoa_result.fval]}
    
    def cplex(self):
        """
        Runs CplexOptimizer supported by qiskit which internally runs cplex's solver.

        :return: result
        """
        result = CplexOptimizer().solve(self.qp)
        return result


from typing import Union
import networkx as nx
from pydantic import BaseModel


class QUBOProblem(BaseModel):
    '''class representing a Maximum Independent Set problem to feed the inputs to the solver
    :type num_variables: int
    :param num_variables: Number of variables in the QUBO

    :type qubo_dict: Optional[dict]
    :param qubo_dict: QUBO to be passed in the form of a dictionary. Every key is an edge represented by tuples of node numbers, and every value is the weight of that edge.

    :type max_evaluations: int
    :param max_evaluations: Number of maximum evaluations to be run to search the solution.
    '''
    class Config:
        arbitrary_types_allowed = True
    num_variables: int = 100
    qubo_dict: dict = None
    max_evaluations: int = 1000
    problem_type: str = "QUBO"


from typing import Union
import networkx as nx
from pydantic import BaseModel
from typing import Optional

class MaximumCutProblem(BaseModel):
    '''class representing a Maximum-Cut problem to feed the inputs to the solver

    Objective function:

    :type num_nodes: int
    :param num_nodes: Number of nodes in the problem graph

    :type graph: Optional[Union[dict, networkx.Graph]]
    :param graph: Graph to be solved for maximum cut. If the graph is not passed the problem method generates a random graph as per given num_nodes.

    :type penalty: int
    :param penalty: penalty term used in the QUBO formulation.

    :type max_evaluations: int
    :param max_evaluations: Number of maximum evaluations to be run to search the solution.

    :type load_gset: Optional[str]
    :param load_gset: Specify one among the G1,...,G77 Stanford Gset dataset or keep it none to pass a graph
    or generate a graph.
    '''
    class Config:
        arbitrary_types_allowed = True

    num_nodes: int = 100
    graph: Optional[Union[dict, nx.Graph]] = None
    penalty: int = 2
    max_evaluations: int = 1000
    load_gset: Optional[str] = None
    problem_type: str = "maximumCut"
